#define LINT_ARGS 1      /* Enable Compiler Parameter Checking */
#include <stdio.h>
#include <math.h>

/*

Reads groups of means and associated covariance matrices,
and output natural log of means, with corresponding
transformation of covariances.

*/


/* Data sizes... */

#define MAX_MEANS 1000
#define MAX_COVNS 250000
#define MAX_GRPS  1000


/* Static workspace... */

double means_acc [MAX_MEANS] ;
double covns_acc [MAX_COVNS] ;

double *covn_ptrs [MAX_MEANS] ;



int init_arrays (mean_grp, covn_grp, num_grps, grp_size)

double  *mean_grp [MAX_GRPS] ;
double **covn_grp [MAX_GRPS] ;
int num_grps, grp_size ;

/*

  Initialise pointer arrays so that 'mean_grp' looks like an array of 
  linear arrays and 'covn_grp' looks like and array of triangular arrays.

  Also initialise data to zeros.

*/

{
  {
    int num_means = num_grps * grp_size ;

    if (num_grps > MAX_GRPS)
    {
      printf ("\nERRROR: Too many groups of observables\n") ;
      return 0 ;
    }
    if (num_means > MAX_MEANS)
    {
      printf ("\nERRROR: Too many observables\n") ;
      return 0 ;
    }
    if (((num_means * (num_means + 1)) / 2) > MAX_COVNS)
    {
      printf ("\nERROR: Too many covariances\n") ;
      return 0 ;
    }
  }

  {
    /*

    Initialise pointer arrays...

    double  *mean_grp [MAX_GRPS] ;
    double **covn_grp [MAX_GRPS] ;

    and (static)

    double *covn_ptrs [MAX_MEANS] ;

    */

    int g, i ;
    double *mean_addr = means_acc ;
    double *covn_addr = covns_acc ;
    double **covn_ptr_addr = covn_ptrs ;

    for (g = 0; g < num_grps; g++)
    {
      mean_grp [g] = mean_addr ;
      covn_grp [g] = covn_ptr_addr ;
      for (i = 0; i < grp_size; i++)
      {
        covn_grp [g] [i] = covn_addr ;
        mean_addr     ++ ;
        covn_ptr_addr ++ ;
        covn_addr     += i + 1 ;
      }
    }
  }

  {

    /* Finally, zero arrays */

    int g, i, j ;

    for (g = 0; g < num_grps; g++)
      for (i = 0; i < grp_size; i++)
      {
        mean_grp [g] [i] = 0.0 ;
        for (j = 0; j <= i; j++)
          covn_grp [g] [i] [j] = 0.0 ;
      }
  }

  return 1 ;
}

void input_group (means, covns, grp_size) 

double  *means ;
double **covns ;
int grp_size ;

/*

  Read group of means and covariances.

*/

{
  int i, j ;

  for (i = 0; i < grp_size; i++)
    scanf ("%lf", &(means [i])) ;

  for (i = 0; i < grp_size; i++)
    for (j = 0; j <= i; j++)
      scanf ("%lf", &(covns [i] [j])) ;
}

void process_group (means, covns, grp_size) 

double  *means ;
double **covns ;
int grp_size ;

/*

  Do transformation on group of means and covariances.

*/

{
  int i, j ;

  for (i = 0; i < grp_size; i++)
    for (j = 0; j <= i; j++)
      covns [i] [j] /= means [i] * means [j] ;

  for (i = 0; i < grp_size; i++)
    if (means [i] > 0.0)
      means [i] = log (means [i]) ;
    else
      means [i] = -9999.0 ;
}

void output_group (means, covns, grp_size) 

double  *means ;
double **covns ;
int grp_size ;

/*

  Pretty-print group of means and covariances (with errors added).

*/

{
  if (grp_size == 1)
    printf ("%13f %13f %13f\n", means [0], sqrt (covns [0] [0]),
                                                 covns [0] [0]) ;
  else
  {
    int i, j, k ;

    for (i = 0; i < grp_size; i++)
      printf ("%13f %13f\n", means [i], sqrt (covns [i] [i])) ;

    printf ("\n") ;

    k = 0 ;
    for (i = 0; i < grp_size; i++)
      for (j = 0; j <= i; j++)
      {
        printf ("%13e ", covns [i] [j]) ;
        k++ ;
        if (k == 5)
        {
          printf ("\n") ;
          k = 0 ;
        }
      }

    printf ("\n\n") ;
  }
}

void main ()
{
  int num_grps, grp_size, num_samples ;
  double  *mean_grp [MAX_GRPS] ;
  double **covn_grp [MAX_GRPS] ;

  read_header (&num_grps, &grp_size, &num_samples) ;

  if (init_arrays (mean_grp, covn_grp, num_grps, grp_size))
  {
    int g ;

    for (g = 0; g < num_grps; g++)
    {
      input_group   (mean_grp [g], covn_grp [g], grp_size) ;
      process_group (mean_grp [g], covn_grp [g], grp_size) ;
      output_group  (mean_grp [g], covn_grp [g], grp_size) ;
    }
  }
}

/*
--
--     End of File.
--
*/
